use crate::{
    data::{
        id::ID,
        record::{Field, Record},
    },
    error::{ErrorType, MTTError},
};
use std::{collections::HashMap, fmt};

#[derive(Clone, Debug)]
pub enum TBLError {
    DuplicateField(String),
    InvalidField(String),
    MissingField(String),
}

impl fmt::Display for TBLError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            TBLError::DuplicateField(data) => write!(f, "'{}' already exists", data),
            TBLError::InvalidField(data) => write!(f, "'{}' invalid field name", data),
            TBLError::MissingField(data) => write!(f, "'{}' field is missing", data),
        }
    }
}

#[cfg(test)]
mod errors {
    use super::*;

    #[test]
    fn duplicate_field() {
        let data = "something";
        let err = TBLError::DuplicateField(data.to_string());
        assert_eq!(err.to_string(), format!("'{}' already exists", data));
    }

    #[test]
    fn invalid_field() {
        let data = "something";
        let err = TBLError::InvalidField(data.to_string());
        assert_eq!(err.to_string(), format!("'{}' invalid field name", data));
    }

    #[test]
    fn missing_field() {
        let data = "something";
        let err = TBLError::MissingField(data.to_string());
        assert_eq!(err.to_string(), format!("'{}' field is missing", data));
    }
}

enum FieldType {
    ID,
}

struct FieldDef;

impl FieldDef {
    fn new() -> Self {
        Self {}
    }

    fn get_type(&self) -> FieldType {
        FieldType::ID
    }
}

pub struct Table {
    fields: HashMap<String, FieldDef>,
}

impl Table {
    pub fn new() -> Self {
        Self {
            fields: HashMap::new(),
        }
    }

    fn description(&self) -> &HashMap<String, FieldDef> {
        &self.fields
    }

    fn add_field(&mut self, name: &str, field_type: FieldType) -> Result<(), MTTError> {
        match self.fields.get(name) {
            Some(_) => {
                let err = TBLError::DuplicateField(name.to_string());
                Err(err.into())
            }
            None => {
                self.fields.insert(name.to_string(), FieldDef::new());
                Ok(())
            }
        }
    }

    fn add_record(&mut self, rec: Record) -> Result<Record, MTTError> {
        let mut holder = rec.clone();
        for (key, _) in self.fields.iter() {
            match holder.remove(key.into()) {
                Some(_) => {}
                None => {
                    let err = TBLError::MissingField(key.to_string());
                    return Err(err.into());
                }
            }
        }
        if holder.len() > 0 {
            for (key, _) in holder.iter() {
                let err = TBLError::InvalidField(key.to_string());
                return Err(err.into());
            }
        }
        Ok(rec)
    }
}

#[cfg(test)]
mod table {
    use super::*;

    #[test]
    fn new_table() {
        let tbl = Table::new();
        let data = tbl.description();
        assert_eq!(data.len(), 0);
    }

    #[test]
    fn add_field() {
        let mut tbl = Table::new();
        let field_name = "something ";
        tbl.add_field(field_name, FieldType::ID);
        let data = tbl.description();
        assert_eq!(data.len(), 1);
        match data.get(field_name) {
            Some(field_info) => match field_info.get_type() {
                FieldType::ID => {}
                _ => unreachable!("incorrect field type"),
            },
            None => unreachable!("should return field definition"),
        }
    }

    #[test]
    fn error_on_duplicate_name() {
        let mut tbl = Table::new();
        let name = "one";
        tbl.add_field(name, FieldType::ID);
        match tbl.add_field(name, FieldType::ID) {
            Ok(_) => unreachable!(" Should not duplicates."),
            Err(err) => match err.get_code() {
                ErrorType::TBLErr(data) => match data {
                    TBLError::DuplicateField(fname) => assert_eq!(fname, name),
                    _ => unreachable!("Should have been a dupluicate field error"),
                },
                _ => unreachable!("should produce a duplicate name error"),
            },
        }
    }

    #[test]
    fn add_record() {
        let mut tbl = Table::new();
        let name = "id";
        let field_data = ID::random();
        tbl.add_field(name, FieldType::ID).unwrap();
        let mut rec = Record::new();
        rec.add(name, field_data.clone());
        let result = tbl.add_record(rec).unwrap();
        assert_eq!(
            result.get(name).unwrap().to_string(),
            field_data.to_string()
        );
    }

    #[test]
    fn add_record_incorrect_field_name() {
        let mut tbl = Table::new();
        let name = "id";
        let field_data = ID::random();
        let mut rec = Record::new();
        rec.add(name, field_data.clone());
        match tbl.add_record(rec) {
            Ok(_) => unreachable!("should have produced an error"),
            Err(err) => match err.get_code() {
                ErrorType::TBLErr(error) => match error {
                    TBLError::InvalidField(result) => assert_eq!(result, name),
                    _ => unreachable!("should have been invalid field name"),
                },
                _ => unreachable!("should have been a table error"),
            },
        }
    }

    #[test]
    fn add_record_missing_field() {
        let mut tbl = Table::new();
        let present = "here";
        let missing = "not";
        tbl.add_field(present, FieldType::ID);
        tbl.add_field(missing, FieldType::ID);
        let mut rec = Record::new();
        rec.add(present, ID::random());
        match tbl.add_record(rec) {
            Ok(_) => unreachable!("should have produced an error"),
            Err(err) => match err.get_code() {
                ErrorType::TBLErr(error) => match error {
                    TBLError::MissingField(result) => assert_eq!(result, missing),
                    _ => unreachable!("should have been missing field"),
                },
                _ => unreachable!("should have been a table error"),
            },
        }
    }
}
